from gitlabform.configuration.case_insensitivity import (
    ConfigurationCaseInsensitiveProjectsAndGroups,
)


# note that we are NOT using mixins here, but only the most advanced subclass
class Configuration(ConfigurationCaseInsensitiveProjectsAndGroups):
    pass



import argparse
import logging
import logging.config
import sys
import textwrap
import traceback
from logging import debug

import cli_ui
from cli_ui import warning, error, fatal

from gitlabform import EXIT_INVALID_INPUT, EXIT_PROCESSING_ERROR
from gitlabform.configuration.core import (
    ConfigFileNotFoundException,
    ConfigInvalidException,
)
from gitlabform.filter import NonEmptyConfigsProvider
from gitlabform.gitlab import GitLab
from gitlabform.gitlab.core import TestRequestFailedException
from gitlabform.input import GroupsAndProjectsProvider
from gitlabform.output import EffectiveConfiguration
from gitlabform.processors.group import GroupProcessors
from gitlabform.processors.project import ProjectProcessors
from gitlabform.transform import AccessLevelsTransformer
from gitlabform.ui import (
    info_group_count,
    info_project_count,
    show_version,
    show_summary,
    show_header,
)


class Formatter(
    argparse.ArgumentDefaultsHelpFormatter,
    argparse.RawTextHelpFormatter,
):
    pass


class GitLabForm(object):
    def __init__(self, include_archived_projects=True, target=None, config_string=None):

        if target and config_string:
            # this mode is basically only for testing

            self.target = target
            self.config_string = config_string
            self.verbose = True
            self.debug = True
            self.strict = True
            self.start_from = 1
            self.start_from_group = 1
            self.noop = False
            self.output_file = None
            self.skip_version_check = True
            self.include_archived_projects = include_archived_projects
            self.just_show_version = False
            self.terminate_after_error = True
            self.only_sections = "all"

            self.configure_output(tests=True)
        else:
            # normal mode

            (
                self.target,
                self.config,
                self.verbose,
                self.debug,
                self.strict,
                self.start_from,
                self.start_from_group,
                self.noop,
                self.output_file,
                self.skip_version_check,
                self.include_archived_projects,
                self.just_show_version,
                self.terminate_after_error,
                self.only_sections,
            ) = self.parse_args()

            self.configure_output()

            show_version(self.skip_version_check)
            if self.just_show_version:
                sys.exit(0)

            if not self.target:
                fatal(
                    "target parameter is required.",
                    exit_code=EXIT_INVALID_INPUT,
                )

        self.access_levels_transformer = AccessLevelsTransformer

        self.gitlab, self.configuration = self.initialize_configuration_and_gitlab()

        self.group_processors = GroupProcessors(
            self.gitlab, self.configuration, self.strict
        )
        self.project_processors = ProjectProcessors(
            self.gitlab, self.configuration, self.strict
        )
        self.groups_and_projects_provider = GroupsAndProjectsProvider(
            self.gitlab,
            self.configuration,
            self.include_archived_projects,
        )

        self.non_empty_configs_provider = NonEmptyConfigsProvider(
            self.configuration, self.group_processors, self.project_processors
        )

    @staticmethod
    def parse_args():

        parser = argparse.ArgumentParser(
            description=textwrap.dedent(
                f"""
            🏗 Specialized "configuration as a code" tool for GitLab projects, groups and more
            using hierarchical configuration written in YAML.

            Exits with code:
              * 0 - on success,
              * {EXIT_INVALID_INPUT} - on invalid input errors (f.e. bad syntax in the config file) ~ "it's your fault". 😅
              * {EXIT_PROCESSING_ERROR} - if there were backend processing errors (f.e. when requests to GitLab fail) ~ "it's not your fault". 😎
            """
            ),
            formatter_class=Formatter,
        )

        parser.add_argument(
            "target",
            nargs="?",
            help='Project name in "group/project" format \n'
            "OR a single group name \n"
            'OR "ALL_DEFINED" to run for all groups and projects defined the config \n'
            'OR "ALL" to run for all projects that you have access to',
        )

        parser.add_argument(
            "-V",
            "--version",
            dest="just_show_version",
            action="store_true",
            help="show version and exit",
        )

        parser.add_argument(
            "-k",
            "--skip-version-check",
            dest="skip_version_check",
            action="store_true",
            help="Skips checking if the latest version is used",
        )

        parser.add_argument(
            "-c", "--config", default="config.yml", help="config file path and filename"
        )

        verbosity_args = parser.add_mutually_exclusive_group()

        verbosity_args.add_argument(
            "-v", "--verbose", action="store_true", help="verbose output"
        )

        verbosity_args.add_argument(
            "-d",
            "--debug",
            action="store_true",
            help="debug output (!!! WARNING !!!: sensitive data and secrets may"
            " be printed in this mode - all the data sent to GitLab API"
            " will be printed in plain-text.)",
        )

        parser.add_argument(
            "-s",
            "--strict",
            action="store_true",
            help="stop on missing branches and tags",
        )

        parser.add_argument(
            "-n",
            "--noop",
            dest="noop",
            action="store_true",
            help="run in no-op (dry run) mode",
        )

        parser.add_argument(
            "-o",
            "--output-file",
            dest="output_file",
            default=None,
            help="name/path of a file to write the effective configs to"
            " (!!! WARNING !!!: if your config contains sensitive data or secrets, then this file will also"
            " contain them.)",
        )

        parser.add_argument(
            "-a",
            "--include-archived-projects",
            dest="include_archived_projects",
            action="store_true",
            help="Includes processing projects that are archived",
        )

        parser.add_argument(
            "-t",
            "--terminate",
            dest="terminate_after_error",
            action="store_true",
            help=f"exit with {EXIT_PROCESSING_ERROR} after the first group/project processing error."
            f" (default: process all the requested groups/projects and skip the failing ones."
            f" At the end, if there were any groups/projects, exit with {EXIT_PROCESSING_ERROR}.)",
        )

        parser.add_argument(
            "-sf",
            "--start-from",
            dest="start_from",
            default=1,
            type=int,
            help="start processing projects from the given one"
            ' (as numbered by "x/y Processing group/project" messages)',
        )

        parser.add_argument(
            "-sfg",
            "--start-from-group",
            dest="start_from_group",
            default=1,
            type=int,
            help="start processing groups from the given one "
            '(as numbered by "x/y Processing group/project" messages)',
        )

        parser.add_argument(
            "-os",
            "--only-sections",
            dest="only_sections",
            default="all",
            type=str,
            help="process only section with these names (comma-delimited)",
        )

        args = parser.parse_args()

        if args.only_sections != "all":
            args.only_sections = args.only_sections.split(",")

        return (
            args.target,
            args.config,
            args.verbose,
            args.debug,
            args.strict,
            args.start_from,
            args.start_from_group,
            args.noop,
            args.output_file,
            args.skip_version_check,
            args.include_archived_projects,
            args.just_show_version,
            args.terminate_after_error,
            args.only_sections,
        )

    def configure_output(self, tests=False):

        # normal mode - print cli_ui.* except debug as verbose
        # verbose mode - print all cli_ui.*, including debug as verbose
        # debug / tests mode - like above + (logging.)debug

        logging.basicConfig()

        if not self.verbose and not self.debug:  # normal
            cli_ui.setup()
            level = (
                logging.FATAL
            )  # de facto disabled as we don't use logging different than debug in this project
        else:
            if (
                self.debug or tests
            ):  # debug (BUT verbose may also be set, that's why we check this first)
                cli_ui.setup(verbose=True)
                level = logging.DEBUG
            elif self.verbose:  # verbose
                cli_ui.setup(verbose=True)
                level = (
                    logging.FATAL
                )  # de facto disabled as we don't use logging different than debug in this project

        logging.getLogger().setLevel(level)
        fmt = logging.Formatter("%(message)s")
        logging.getLogger().handlers[0].setFormatter(fmt)

    def initialize_configuration_and_gitlab(self):

        try:
            if hasattr(self, "config_string"):
                gitlab = GitLab(config_string=self.config_string)
            else:
                gitlab = GitLab(config_path=self.config)
            configuration = gitlab.get_configuration()
            self.access_levels_transformer.transform(configuration)
            return gitlab, configuration
        except ConfigFileNotFoundException as e:
            fatal(
                f"Config file not found at: {e}",
                exit_code=EXIT_INVALID_INPUT,
            )
        except ConfigInvalidException as e:
            fatal(
                f"Invalid config:\n{e.underlying}",
                exit_code=EXIT_INVALID_INPUT,
            )
        except TestRequestFailedException as e:
            fatal(
                f"GitLab test request failed:\n{e.underlying}",
                exit_code=EXIT_PROCESSING_ERROR,
            )

    def run(self):

        projects, groups = show_header(
            self.target,
            self.groups_and_projects_provider,
            self.non_empty_configs_provider,
        )

        group_number = 0
        successful_groups = 0
        failed_groups = {}

        effective_configuration = EffectiveConfiguration(self.output_file)

        for group in groups:

            group_number += 1

            if group_number < self.start_from_group:
                info_group_count(
                    "@",
                    group_number,
                    len(groups),
                    cli_ui.yellow,
                    f"Skipping group {group} as requested to start from {self.start_from_group}...",
                    cli_ui.reset,
                )
                continue

            configuration = self.configuration.get_effective_config_for_group(group)

            effective_configuration.add_placeholder(group)

            info_group_count(
                "@",
                group_number,
                len(groups),
                f"Processing group: {group}",
            )

            try:
                self.group_processors.process_entity(
                    group,
                    configuration,
                    dry_run=self.noop,
                    effective_configuration=effective_configuration,
                    only_sections=self.only_sections,
                )

                successful_groups += 1

            except Exception as e:

                failed_groups[group_number] = group

                trace = traceback.format_exc()
                message = (
                    f"Error occurred while processing group {group}, exception:\n\n{e}"
                )

                if self.terminate_after_error:
                    effective_configuration.write_to_file()
                    error(message)
                    debug(trace)
                    sys.exit(EXIT_PROCESSING_ERROR)
                else:
                    warning(message)
                    debug(trace)
            finally:
                debug(
                    f"@ ({group_number}/{len(groups)}) FINISHED Processing group: {group}"
                )

        project_number = 0
        successful_projects = 0
        failed_projects = {}

        for project_and_group in projects:

            project_number += 1

            if project_number < self.start_from:
                info_project_count(
                    "*",
                    project_number,
                    len(projects),
                    cli_ui.yellow,
                    f"Skipping project {project_and_group} as requested to start from {self.start_from}...",
                    cli_ui.reset,
                )
                continue

            configuration = self.configuration.get_effective_config_for_project(
                project_and_group
            )

            effective_configuration.add_placeholder(project_and_group)

            info_project_count(
                "*",
                project_number,
                len(projects),
                f"Processing project: {project_and_group}",
            )

            try:
                self.project_processors.process_entity(
                    project_and_group,
                    configuration,
                    dry_run=self.noop,
                    effective_configuration=effective_configuration,
                    only_sections=self.only_sections,
                )

                successful_projects += 1

            except Exception as e:

                failed_projects[project_number] = project_and_group

                trace = traceback.format_exc()
                message = f"Error occurred while processing project {project_and_group}, exception:\n\n{e}"

                if self.terminate_after_error:
                    effective_configuration.write_to_file()
                    error(message)
                    debug(trace)
                    sys.exit(EXIT_PROCESSING_ERROR)
                else:
                    warning(message)
                    debug(trace)
            finally:

                debug(
                    f"* ({project_number}/{len(projects)})"
                    f" FINISHED Processing project: {project_and_group}",
                )

        effective_configuration.write_to_file()

        show_summary(
            groups,
            projects,
            successful_groups,
            successful_projects,
            failed_groups,
            failed_projects,
        )



from abc import ABC, abstractmethod

from cli_ui import debug as verbose

from gitlabform.gitlab import GitLab
from gitlabform.output import EffectiveConfiguration
from gitlabform.processors.util.decorators import configuration_to_safe_dict


class AbstractProcessor(ABC):
    def __init__(self, configuration_name: str, gitlab: GitLab):
        self.configuration_name = configuration_name
        self.gitlab = gitlab

    @configuration_to_safe_dict
    def process(
        self,
        project_or_project_and_group: str,
        configuration: dict,
        dry_run: bool,
        effective_configuration: EffectiveConfiguration,
    ):
        if self._section_is_in_config(configuration):
            if configuration.get(f"{self.configuration_name}|skip"):
                verbose(
                    f"Skipping section '{self.configuration_name}' - explicitly configured to do so."
                )
                return
            elif (
                configuration.get("project|archive")
                and self.configuration_name != "project"
            ):
                verbose(
                    f"Skipping section '{self.configuration_name}' - it is configured to be archived."
                )
                return

            if dry_run:
                verbose(
                    f"Processing section '{self.configuration_name}' in dry-run mode."
                )
                self._print_diff(
                    project_or_project_and_group,
                    configuration.get(self.configuration_name),
                )
            else:
                verbose(f"Processing section '{self.configuration_name}'")
                self._process_configuration(project_or_project_and_group, configuration)

            effective_configuration.add_configuration(
                project_or_project_and_group,
                self.configuration_name,
                configuration.get(self.configuration_name),
            )
        else:
            verbose(f"Skipping section '{self.configuration_name}' - not in config.")

    def _section_is_in_config(self, configuration: dict):
        return self.configuration_name in configuration

    @abstractmethod
    def _process_configuration(
        self, project_or_project_and_group: str, configuration: dict
    ):
        pass

    def _print_diff(self, project_or_project_and_group: str, entity_config):
        verbose(f"Diffing for section '{self.configuration_name}' is not supported yet")

    @staticmethod
    def _needs_update(
        entity_in_gitlab: dict,
        entity_in_configuration: dict,
    ):
        # in the configuration we often don't define every key value because we rely on the defaults.
        # that's why GitLab API often returns many more keys than we have in the configuration.

        # so to decide if the entity should be updated:
        # a) we look for ANY settings that are ONLY in configuration,
        # a) we compare the settings that are in both configuration and gitlab,

        keys_only_in_configuration = set(entity_in_configuration.keys()) - set(
            entity_in_gitlab.keys()
        )
        if len(keys_only_in_configuration) > 0:
            return True

        keys_on_both_sides = set(entity_in_configuration.keys()) & set(
            entity_in_gitlab.keys()
        )
        for key in keys_on_both_sides:
            if entity_in_gitlab[key] != entity_in_configuration[key]:
                return True

        return False



from gitlabform.gitlab.projects import GitLabProjects


class GitLabProjectDeployKeys(GitLabProjects):
    def get_deploy_keys(self, project_and_group_name):
        return self._make_requests_to_api(
            "projects/%s/deploy_keys", project_and_group_name
        )

    def post_deploy_key(self, project_and_group_name, deploy_key_in_config):
        # deploy_key_in_config has to be like this:
        # {
        #     'title': title,
        #     'key': key,
        #     'can_push': can_push,
        # }
        # ..as documented at: https://docs.gitlab.com/ce/api/deploy_keys.html#add-deploy-key
        self._make_requests_to_api(
            "projects/%s/deploy_keys",
            project_and_group_name,
            "POST",
            deploy_key_in_config,
            expected_codes=201,
        )

    def put_deploy_key(
        self, project_and_group_name, deploy_key_in_gitlab, deploy_key_in_config
    ):
        # according to docs_new at https://docs.gitlab.com/ee/api/deploy_keys.html#update-deploy-key
        # you only can change key's title and can_push param

        changeable_data = {
            "title": deploy_key_in_config.get("title", None),
            "can_push": deploy_key_in_config.get("can_push", None),
        }

        return self._make_requests_to_api(
            "projects/%s/deploy_keys/%s",
            (project_and_group_name, deploy_key_in_gitlab["id"]),
            "PUT",
            changeable_data,
        )

    def delete_deploy_key(self, project_and_group_name, deploy_key_in_config):
        return self._make_requests_to_api(
            "projects/%s/deploy_keys/%s",
            (project_and_group_name, deploy_key_in_config["id"]),
            method="DELETE",
            expected_codes=[204, 404],
        )

    def get_deploy_key(self, project_and_group_name, id):
        return self._make_requests_to_api(
            "projects/%s/deploy_keys/%s", (project_and_group_name, id), "GET"
        )



import pytest
from deepdiff import DeepDiff

from gitlabform import EXIT_INVALID_INPUT
from gitlabform.configuration import Configuration
from gitlabform.transform import AccessLevelsTransformer


def test__config__with_access_level_names__branches():
    config_yaml = f"""
    projects_and_groups:
      foobar/*:
        branches:
          develop:
            protected: false
          main:
            protected: true
            push_access_level: no access
            merge_access_level: developer
            unprotect_access_level: maintainer
            code_owner_approval_required: true
          branch_protected_from_changes:
            protected: true
            push_access_level: no access
            merge_access_level: no access
            unprotect_access_level: maintainer
      "*":
        branches:
          '*-something':
            protected: true
            push_access_level: no access
            merge_access_level: developer
            unprotect_access_level: maintainer
          allow_to_force_push:
            protected: true
            push_access_level: developer
            merge_access_level: developer
            unprotect_access_level: maintainer
            allow_force_push: true
    """
    configuration = Configuration(config_string=config_yaml)

    AccessLevelsTransformer.transform(configuration)

    config_with_numbers = f"""
    projects_and_groups:
      foobar/*:
        branches:
          develop:
            protected: false
          main:
            protected: true
            push_access_level: 0
            merge_access_level: 30
            unprotect_access_level: 40
            code_owner_approval_required: true
          branch_protected_from_changes:
            protected: true
            push_access_level: 0
            merge_access_level: 0
            unprotect_access_level: 40
      "*":
        branches:
          '*-something':
            protected: true
            push_access_level: 0
            merge_access_level: 30
            unprotect_access_level: 40
          allow_to_force_push:
            protected: true
            push_access_level: 30
            merge_access_level: 30
            unprotect_access_level: 40
            allow_force_push: true
    """
    configuration_with_numbers = Configuration(config_string=config_with_numbers)

    ddiff = DeepDiff(configuration.config, configuration_with_numbers.config)
    assert not ddiff


def test__config__with_access_level_names__group_ldap_links():
    config_yaml = f"""
    projects_and_groups:
      foobar/*:
        group_ldap_links:
          # "provider" field should contain a value that you can find in the GitLab web UI,
          # see https://github.com/gitlabform/gitlabform/issues/261
          devops_are_maintainers:
            provider: "AD"
            cn: "devops"
            group_access: maintainer
          developers_are_developers:
            provider: "AD"
            filter: "(employeeType=developer)"
            group_access: developer
    """
    configuration = Configuration(config_string=config_yaml)

    AccessLevelsTransformer.transform(configuration)

    config_with_numbers = f"""
    projects_and_groups:
      foobar/*:
        group_ldap_links:
          # "provider" field should contain a value that you can find in the GitLab web UI,
          # see https://github.com/gitlabform/gitlabform/issues/261
          devops_are_maintainers:
            provider: "AD"
            cn: "devops"
            group_access: 40
          developers_are_developers:
            provider: "AD"
            filter: "(employeeType=developer)"
            group_access: 30
    """
    configuration_with_numbers = Configuration(config_string=config_with_numbers)

    ddiff = DeepDiff(configuration.config, configuration_with_numbers.config)
    assert not ddiff


def test__config__with_access_level_names__branches_premium_syntax():
    config_yaml = f"""
    projects_and_groups:
      foobar/*:
        branches:
          special:
            protected: true
            allowed_to_push:
              - user: jsmith # you can use usernames...
              - user: bdoe
              - group: another-group # ...or group names (paths)...
            allowed_to_merge:
              - user_id: 15 # ...or user ids, if you know them...
              - access_level: developer # ...or the whole access levels, like in the other syntax
              - group_id: 456 # ...or group ids, if you know them...
            allowed_to_unprotect:
              - access_level: maintainer # ...or the whole access levels, like in the other syntax
    """
    configuration = Configuration(config_string=config_yaml)

    AccessLevelsTransformer.transform(configuration)

    config_with_numbers = f"""
    projects_and_groups:
      foobar/*:
        branches:
          special:
            protected: true
            allowed_to_push:
              - user: jsmith # you can use usernames...
              - user: bdoe
              - group: another-group # ...or group names (paths)...
            allowed_to_merge:
              - user_id: 15 # ...or user ids, if you know them...
              - access_level: 30 # ...or the whole access levels, like in the other syntax
              - group_id: 456 # ...or group ids, if you know them...
            allowed_to_unprotect:
              - access_level: 40 # ...or the whole access levels, like in the other syntax
    """
    configuration_with_numbers = Configuration(config_string=config_with_numbers)

    ddiff = DeepDiff(configuration.config, configuration_with_numbers.config)
    assert not ddiff


def test__config__with_access_level_names__invalid_name():
    config_yaml = f"""
    projects_and_groups:
      foobar/*:
        branches:
          special:
            protected: true
            allowed_to_push:
              - user: jsmith # you can use usernames...
              - user: bdoe
              - group: another-group # ...or group names (paths)...
            allowed_to_merge:
              - user_id: 15 # ...or user ids, if you know them...
              - access_level: developers # <-------------------------- this is invalid, it's plural
              - group_id: 456 # ...or group ids, if you know them...
            allowed_to_unprotect:
              - access_level: maintainer # ...or the whole access levels, like in the other syntax
    """
    configuration = Configuration(config_string=config_yaml)

    with pytest.raises(SystemExit) as e:
        AccessLevelsTransformer.transform(configuration)
    assert e.value.code == EXIT_INVALID_INPUT



from logging import debug
from cli_ui import debug as verbose

from distutils.version import LooseVersion

from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor
from gitlabform.processors.util.decorators import SafeDict
from gitlabform.processors.util.difference_logger import DifferenceLogger


class MergeRequestsProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("merge_requests", gitlab)

    def _process_configuration(self, project_and_group: str, configuration: dict):
        approvals = configuration.get("merge_requests|approvals")
        if approvals:
            verbose(f"Setting approvals settings: {approvals}")
            self.gitlab.post_approvals_settings(project_and_group, approvals)

        approvers = configuration.get("merge_requests|approvers")
        approver_groups = configuration.get("merge_requests|approver_groups")
        remove_other_approval_rules = configuration.get(
            "merge_requests|remove_other_approval_rules"
        )
        # checking if "is not None" allows configs with empty array to work
        if (
            approvers is not None
            or approver_groups is not None
            and approvals
            and "approvals_before_merge" in approvals
        ):
            verbose(f"Setting approvers...")

            approval_rule_name = "Approvers (configured using GitLabForm)"

            # is a rule already configured and just needs updating?
            approval_rule_id = None
            rules = self.gitlab.get_approvals_rules(project_and_group)
            for rule in rules:
                if rule["name"] == approval_rule_name:
                    approval_rule_id = rule["id"]
                else:
                    if remove_other_approval_rules:
                        debug("Deleting extra approval rule '%s'" % rule["name"])
                        self.gitlab.delete_approvals_rule(project_and_group, rule["id"])

            if not approvers:
                approvers = []
            if not approver_groups:
                approver_groups = []

            if approval_rule_id:
                # the rule exists, needs an update
                verbose(
                    f"Updating approvers rule to users {approvers} and groups {approver_groups}"
                )
                self.gitlab.update_approval_rule(
                    project_and_group,
                    approval_rule_id,
                    approval_rule_name,
                    approvals["approvals_before_merge"],
                    approvers,
                    approver_groups,
                )
            else:
                # the rule does not exist yet, let's create it
                verbose(
                    f"Creating approvers rule to users {approvers} and groups {approver_groups}"
                )
                self.gitlab.create_approval_rule(
                    project_and_group,
                    approval_rule_name,
                    approvals["approvals_before_merge"],
                    approvers,
                    approver_groups,
                )

    def _print_diff(self, project_and_group: str, merge_requests: SafeDict):
        approvals = merge_requests.get("approvals")
        if approvals:
            DifferenceLogger.log_diff(
                "Project %s approvals changes" % project_and_group, dict(), approvals
            )



from tests.acceptance import (
    run_gitlabform,
)


class TestDeployKeysAllProjects:
    def test__deploy_key_to_all_projects(self, gitlab, group, project, other_project):
        # noinspection PyPep8
        deploy_key_to_all_projects = """
        projects_and_groups:
          "*":
            deploy_keys:
              foobar:
                key: ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABgQC6OxCCViSjh8QUKNOoGqhUqs4LLDMyq/7DYuvMJu5lXwECWp0wFGoLXzYWCT6WOAP+vccncOrlVfsr9VJzXxR1QZq+p3joW25nWgjEw/HCPI6fnU1vROImzxnvwLS3EEJpy64Jq0FFwjt8vKSuQshPysEBSUTf5t3omb166MGlZ+Y6/tOf/8/3zqmvb8OqNmhUtfwxfE5oX8Z8bBaGrkxHlmYyJ9UBpfeEcFt1GqfiONPgchJJ4OqCJKqd7H4DZOosT64kTqPXhca44EOxiKQviCthv7bO+r7VSFo5TVo60ikq/sTR9ifXnd3B9x3LV1qzHHLlmnP//xkKHIZGxfyhgwtdGNWhEtKPiXUzZv4/48WUJMmtpjznhuEgjnpiJL3x0+vJCStA6WG0MiozBlS80Y4XHbt3X3bvlNSqSo/GpnxlPTUx+Lj/ASI75JDym14+C8RdSFN4iKl5Qjz5xFq4eXke00AahFvjAAV5BT8Qrlyg/cbt1pfWKND1T5Fqh6c=
                title: common_key
                can_push: false
        """
        run_gitlabform(deploy_key_to_all_projects, group)

        deploy_keys1 = gitlab.get_deploy_keys(f"{group}/{project}")
        assert len(deploy_keys1) == 1

        deploy_keys2 = gitlab.get_deploy_keys(f"{group}/{other_project}")
        assert len(deploy_keys2) == 1



import pytest

from tests.acceptance import run_gitlabform, gl


class TestGroupSettings:
    def test__edit_settings(self, gitlab, group):
        settings = gitlab.get_group_settings(group)
        assert settings["description"] != "foobar"

        edit_group_settings = f"""
        projects_and_groups:
          {group}/*:
            group_settings:
              description: foobar
        """

        run_gitlabform(edit_group_settings, group)

        settings = gitlab.get_group_settings(group)
        assert settings["visibility"] == "internal"

    def test__no_edit_needed(self, gitlab, group):
        settings = gitlab.get_group_settings(group)
        current_value = settings["path"]

        edit_group_settings = f"""
            projects_and_groups:
              {group}/*:
                group_settings:
                  path: {current_value}
            """

        run_gitlabform(edit_group_settings, group)

        settings = gitlab.get_group_settings(group)
        assert settings["path"] == current_value

    @pytest.mark.skipif(
        gl.has_no_license(), reason="this test requires a GitLab license (Paid/Trial)"
    )
    def test__edit_new_setting_premium(self, gitlab, project, group):

        project_id = gitlab._get_project_id(f"{group}/{project}")

        settings = gitlab.get_group_settings(group)
        assert "file_template_project_id" not in settings

        edit_group_settings = f"""
        projects_and_groups:
          {group}/*:
            group_settings:
              file_template_project_id: {project_id}
        """

        run_gitlabform(edit_group_settings, group)

        settings = gitlab.get_group_settings(group)
        # the type returned by the API is int, but in the _get_project_id we return str,
        # so we need a cast here
        assert settings["file_template_project_id"] == int(project_id)



from collections import defaultdict
from typing import DefaultDict

from abc import ABC

# f.e. bad syntax in the config file ~= "it's your fault" 😅
EXIT_INVALID_INPUT = 1
# f.e. when requests to GitLab fail ~= "it's not your fault" 😎
EXIT_PROCESSING_ERROR = 2


class Entities(ABC):
    def __init__(self, name: str):
        self.requested: set = set()
        self.omitted: DefaultDict[str, set] = defaultdict(set)
        self.effective: list = []
        self.recalculate_effective = True
        self.name = name

    def add_requested(self, more_requested: list):
        self.requested = self.requested | set(more_requested)
        self.recalculate_effective = True

    def add_omitted(self, reason: str, more_omitted: list):
        self.omitted[reason] = self.omitted[reason] | set(more_omitted)
        self.recalculate_effective = True

    def get_omitted(self, reason: str) -> list:
        return sorted(self.omitted[reason])

    def any_omitted(self) -> bool:
        for reason in self.omitted:
            if len(self.omitted[reason]) > 0:
                return True
        return False

    def get_effective(self) -> list:
        if self.recalculate_effective:
            self.effective = self.requested
            for reason in self.omitted:
                self.effective = self.effective - self.omitted[reason]
            self.effective = sorted(self.effective)
        return self.effective


class Groups(Entities):
    def __init__(self):
        super().__init__("groups")


class Projects(Entities):
    def __init__(self):
        super().__init__("projects")



import json
from typing import Any
from urllib.error import URLError

import luddite
import pkg_resources
import sys
from cli_ui import debug as verbose
from cli_ui import (
    message,
    info,
    info_1,
    error,
    fatal,
    reset,
    green,
    purple,
    blue,
    red,
    yellow,
    Symbol,
    Token,
)
from packaging import version as packaging_version

from gitlabform import EXIT_PROCESSING_ERROR, EXIT_INVALID_INPUT, Entities


def show_version(skip_version_check: bool):
    local_version = pkg_resources.get_distribution("gitlabform").version

    tower_crane = Symbol("🏗", "")
    tokens_to_show = [
        reset,
        tower_crane,
        " GitLabForm version:",
        blue,
        local_version,
        reset,
    ]

    message(*tokens_to_show, end="")

    if skip_version_check:
        # just print end of the line
        print()
    else:
        try:
            latest_version = luddite.get_version_pypi("gitlabform")
        except URLError as e:
            # end the line with current version
            print()
            error(f"Checking latest version failed:\n{e}")
            return

        if local_version == latest_version:
            happy = Symbol("😊", "")
            tokens_to_show = [
                "= the latest stable ",
                happy,
            ]
        elif packaging_version.parse(local_version) < packaging_version.parse(
            latest_version
        ):
            sad = Symbol("😔", "")
            tokens_to_show = [
                "= outdated ",
                sad,
                f", please update! (the latest stable is ",
                blue,
                latest_version,
                reset,
                ")",
            ]
        else:
            excited = Symbol("🤩", "")
            tokens_to_show = [
                "= pre-release ",
                excited,
                f" (the latest stable is ",
                blue,
                latest_version,
                reset,
                ")",
            ]

        message(*tokens_to_show, sep="")


def show_header(
    target,
    groups_and_projects_provider,
    non_empty_configs_provider,
):
    if target == "ALL":
        info(">>> Getting ALL groups and projects...")
    elif target == "ALL_DEFINED":
        info(">>> Getting ALL_DEFINED groups and projects...")
    else:
        info(">>> Getting requested groups/projects...")

    groups, projects = groups_and_projects_provider.get_groups_and_projects(target)

    if len(groups.get_effective()) == 0 and len(projects.get_effective()) == 0:
        if target == "ALL":
            error_message = "GitLab has no projects and groups!"
        elif target == "ALL_DEFINED":
            error_message = (
                "Configuration does not have any groups or projects defined!"
            )
        else:
            error_message = f"Project or group {target} cannot be found in GitLab!"
        fatal(
            error_message,
            exit_code=EXIT_INVALID_INPUT,
        )

    (
        groups,
        projects,
    ) = non_empty_configs_provider.omit_groups_and_projects_with_empty_configs(
        groups, projects
    )

    show_input_entities(groups)
    show_input_entities(projects)

    return projects.get_effective(), groups.get_effective()


def show_input_entities(entities: Entities):
    info_1(f"# of {entities.name} to process: {len(entities.get_effective())}")

    entities_omitted = ""
    entities_verbose = f"{entities.name}: {entities.get_effective()}"
    if entities.any_omitted():
        entities_omitted += f"(# of omitted {entities.name} -"
        first = True
        for reason in entities.omitted:
            if len(entities.omitted[reason]) > 0:
                if not first:
                    entities_omitted += ","
                entities_omitted += f" {reason}: {len(entities.omitted[reason])}"
                entities_verbose += f"\nomitted {entities.name} - {reason}: {entities.get_omitted(reason)}"
                first = False
        entities_omitted += ")"

    if entities_omitted:
        info_1(entities_omitted)

    verbose(entities_verbose)


def show_summary(
    groups_with_non_empty_configs: list,
    projects_with_non_empty_configs: list,
    successful_groups: int,
    successful_projects: int,
    failed_groups: dict,
    failed_projects: dict,
):
    if (
        len(groups_with_non_empty_configs) > 0
        or len(projects_with_non_empty_configs) > 0
    ):
        info_1(f"# of groups processed successfully: {successful_groups}")
        info_1(f"# of projects processed successfully: {successful_projects}")

    if len(failed_groups) > 0:
        info_1(red, f"# of groups failed: {len(failed_groups)}", reset)
        for group_number in failed_groups.keys():
            info_1(
                red,
                f"Failed group {group_number}: {failed_groups[group_number]}",
                reset,
            )
    if len(failed_projects) > 0:
        info_1(
            red,
            f"# of projects failed: {len(failed_projects)}",
            reset,
        )
        for project_number in failed_projects.keys():
            info_1(
                red,
                f"Failed project {project_number}: {failed_projects[project_number]}",
                reset,
            )

    if len(failed_groups) > 0 or len(failed_projects) > 0:
        sys.exit(EXIT_PROCESSING_ERROR)
    elif successful_groups > 0 or successful_projects > 0:
        shine = Symbol("✨", "!!!")
        info_1(
            green,
            f"All requested groups/projects processed successfully!",
            reset,
            shine,
        )
    else:
        info_1(
            yellow,
            "Nothing to do.",
            reset,
        )


def info_group_count(prefix, i: int, n: int, *rest: Token, **kwargs: Any) -> None:
    info_count(purple, prefix, i, n, *rest, **kwargs)


def info_project_count(prefix, i: int, n: int, *rest: Token, **kwargs: Any) -> None:
    info_count(green, prefix, i, n, *rest, **kwargs)


def info_count(color, prefix, i: int, n: int, *rest: Token, **kwargs: Any) -> None:
    num_digits = len(str(n))
    counter_format = "(%{}d/%d)".format(num_digits)
    counter_str = counter_format % (i, n)
    info(color, prefix, reset, counter_str, reset, *rest, **kwargs)


def to_str(a_dict: dict) -> str:
    # arguably the most readable form of a dict in a single line
    # is JSON with sorted keys
    return json.dumps(a_dict, sort_keys=True)



from logging import debug
from cli_ui import debug as verbose
from cli_ui import fatal

import abc
from typing import Callable, List

from gitlabform import EXIT_INVALID_INPUT
from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor
from gitlabform.processors.defining_keys import AbstractKey


class MultipleEntitiesProcessor(AbstractProcessor, metaclass=abc.ABCMeta):
    def __init__(
        self,
        configuration_name: str,
        gitlab: GitLab,
        list_method_name: str,
        add_method_name: str,
        delete_method_name: str,
        defining: AbstractKey,
        required_to_create_or_update: AbstractKey,
        edit_method_name=None,
    ):
        super().__init__(configuration_name, gitlab)
        self.list_method: Callable = getattr(self.gitlab, list_method_name)
        self.add_method: Callable = getattr(self.gitlab, add_method_name)
        self.delete_method: Callable = getattr(self.gitlab, delete_method_name)
        self.defining: AbstractKey = defining
        self.required_to_create_or_update: AbstractKey = required_to_create_or_update

        if edit_method_name:
            self.edit_method: Callable = getattr(self.gitlab, edit_method_name)
        else:
            self.edit_method = None

    def _can_proceed(self, project_or_group: str, configuration: dict):
        return True

    def _process_configuration(self, project_or_group: str, configuration: dict):

        if not self._can_proceed(project_or_group, configuration):
            return

        entities_in_configuration = configuration[self.configuration_name]

        # TODO: move/convert this to a configuration validation phase
        self._find_duplicates(project_or_group, entities_in_configuration)

        entities_in_gitlab = self.list_method(project_or_group)
        debug(f"{self.configuration_name} BEFORE: ^^^")

        # 1. group entities into 3 groups:
        # a) only in configuration,
        # b) in both configuration and gitlab (matching defining keys),
        # c) only in gitlab,

        # 2. if "enforce", then delete c)
        # TODO: implement this.

        # 3. update b), if needed (or delete them if marked as "delete")

        # 4. add a) (or do nothing if marked as "delete")

        for entity_name, entity_config in entities_in_configuration.items():

            entity_in_gitlab = self._is_in_gitlab(entity_config, entities_in_gitlab)
            if entity_in_gitlab:
                if entity_config.get("delete", False):
                    self._validate_required_to_delete(
                        project_or_group, entity_name, entity_config
                    )
                    verbose(
                        f"Deleting {entity_name} of {self.configuration_name} in {project_or_group}"
                    )
                    self.delete_method(project_or_group, entity_in_gitlab)
                elif self._needs_update(entity_in_gitlab, entity_config):
                    self._validate_required_to_create_or_update(
                        project_or_group, entity_name, entity_config
                    )
                    if self.edit_method:
                        verbose(
                            f"Editing {entity_name} of {self.configuration_name} in {project_or_group}"
                        )
                        self.edit_method(
                            project_or_group, entity_in_gitlab, entity_config
                        )
                        debug(f"{self.configuration_name} AFTER: ^^^")
                    else:
                        verbose(
                            f"Recreating {entity_name} of {self.configuration_name} in {project_or_group}"
                        )
                        self.delete_method(project_or_group, entity_in_gitlab)
                        self.add_method(project_or_group, entity_config)
                        debug(f"{self.configuration_name} AFTER: ^^^")
                else:
                    verbose(
                        f"{entity_name} of {self.configuration_name} in {project_or_group} doesn't need an update."
                    )
            else:
                if entity_config.get("delete", False):
                    verbose(
                        f"{entity_name} of {self.configuration_name} in {project_or_group} already doesn't exist."
                    )
                else:
                    self._validate_required_to_create_or_update(
                        project_or_group, entity_name, entity_config
                    )
                    verbose(
                        f"Adding {entity_name} of {self.configuration_name} in {project_or_group}"
                    )
                    self.add_method(project_or_group, entity_config)
                    debug(f"{self.configuration_name} AFTER: ^^^")

    def _find_duplicates(self, project_or_group: str, entities_in_configuration: dict):
        for first_key, first_value in entities_in_configuration.items():
            for second_key, second_value in entities_in_configuration.items():
                if first_key != second_key:
                    if self.defining.matches(first_value, second_value):
                        fatal(
                            f"Entities {first_key} and {second_key} in {self.configuration_name} for {project_or_group}"
                            f" are the same in terms of their defining keys: {self.defining.explain()}",
                            exit_code=EXIT_INVALID_INPUT,
                        )

    def _validate_required_to_create_or_update(
        self, project_or_group: str, entity_name: str, entity_dict: dict
    ):
        if not self.required_to_create_or_update.contains(entity_dict):
            fatal(
                f"Entity {entity_name} in {self.configuration_name} for {project_or_group}"
                f" doesn't have some of its keys required to create or update:"
                f" {self.required_to_create_or_update.explain()}",
                exit_code=EXIT_INVALID_INPUT,
            )

    def _validate_required_to_delete(
        self, project_or_group: str, entity_name: str, entity_dict: dict
    ):
        if not self.defining.contains(entity_dict):
            fatal(
                f"Entity {entity_name} in {self.configuration_name} for {project_or_group}"
                f" doesn't have some of its defining keys required to delete it: {self.defining.explain()}",
                exit_code=EXIT_INVALID_INPUT,
            )

    def _is_in_gitlab(
        self, entity_in_configuration: dict, entities_in_gitlab: List[dict]
    ):
        for entity_in_gitlab in entities_in_gitlab:
            if self.defining.matches(entity_in_gitlab, entity_in_configuration):
                return entity_in_gitlab

        return False



import os
import textwrap
from abc import ABC
from copy import deepcopy
from logging import debug
from pathlib import Path
from ruamel.yaml.scalarstring import ScalarString
from types import SimpleNamespace

from cli_ui import debug as verbose
from cli_ui import fatal
from mergedeep import merge
from yamlpath.common import Parsers
from yamlpath.wrappers import ConsolePrinter

from gitlabform import EXIT_INVALID_INPUT
from ruamel.yaml.comments import CommentedMap


class ConfigurationCore(ABC):
    def __init__(self, config_path=None, config_string=None):

        if config_path and config_string:
            fatal(
                "Please initialize with either config_path or config_string, not both.",
                exit_code=EXIT_INVALID_INPUT,
            )
        try:
            if config_string:
                self.config = self._parse_yaml(config_string, config_string=True)
                self.config_dir = "."
            else:  # maybe config_path
                config_path = self._get_config_path(config_path)
                self.config = self._parse_yaml(config_path, config_string=False)
                self.config_dir = os.path.dirname(config_path)

                # below checks are only needed in the non-test mode, when the config is read from file

                if self.config.get("example_config"):
                    fatal(
                        "Example config detected, aborting.\n"
                        "Haven't you forgotten to use `-c <config_file>` parameter?\n"
                        "If you created your config based on the example config.yml,"
                        " then please remove 'example_config' key.",
                        exit_code=EXIT_INVALID_INPUT,
                    )

                if self.config.get("config_version", 1) != 3:
                    fatal(
                        "This version of GitLabForm requires 'config_version: 3' entry in the config. "
                        "This ensures that if the application behavior changes in a backward-incompatible way,"
                        " you won't apply unwanted configuration to your GitLab instance.\n"
                        "Please follow this guide: https://gitlabform.github.io/gitlabform/upgrade/\n",
                        exit_code=EXIT_INVALID_INPUT,
                    )

                # we are NOT checking for the existence of non-empty 'projects_and_groups' key here
                # as it would break using GitLabForm as a library

        except (FileNotFoundError, IOError):
            raise ConfigFileNotFoundException(config_path)

        except Exception as e:
            raise ConfigInvalidException(e)

    @staticmethod
    def _get_config_path(config_path):

        if "APP_HOME" in os.environ:
            # using this env var should be considered unofficial, we need this temporarily
            # for backwards compatibility. support for it may be removed without notice, do not use it!
            config_path = os.path.join(os.environ["APP_HOME"], "config.yml")
        elif not config_path:
            # this case is only meant for using gitlabform as a library
            config_path = os.path.join(str(Path.home()), ".gitlabform", "config.yml")
        elif config_path in [os.path.join(".", "config.yml"), "config.yml"]:
            # provided points to config.yml in the app current working dir
            config_path = os.path.join(os.getcwd(), "config.yml")

        return config_path

    @staticmethod
    def _parse_yaml(source: str, config_string: bool):

        logging_args = SimpleNamespace(quiet=False, verbose=False, debug=False)
        log = ConsolePrinter(logging_args)

        yaml = Parsers.get_yaml_editor()

        # for better backward compatibility with PyYAML (that supports only YAML 1.1) used in the previous
        # GitLabForm versions, let's force ruamel.yaml to use YAML version 1.1 by default too
        yaml.version = (1, 1)

        if config_string:
            config_string = textwrap.dedent(source)
            verbose("Reading config from the provided string.")
            (yaml_data, doc_loaded) = Parsers.get_yaml_data(
                yaml, log, config_string, literal=True
            )
        else:
            config_path = source
            verbose(f"Reading config from file: {config_path}")
            (yaml_data, doc_loaded) = Parsers.get_yaml_data(yaml, log, config_path)

        if doc_loaded:
            debug("Config parsed successfully as YAML.")
        else:
            # an error message has already been printed via ConsolePrinter
            exit(EXIT_INVALID_INPUT)

        return yaml_data

    def get(self, path, default=None):
        """
        :param path: "path" to given element in YAML file, for example for:

        group_settings:
          sddc:
            deploy_keys:
              qa_puppet:
                key: some key...
                title: some title...
                can_push: false

        ..a path to a single element array ['qa_puppet'] will be: "group_settings|sddc|deploy_keys".

        To get the dict under it use: get("group_settings|sddc|deploy_keys")

        :param default: the value to return if the key is not found. The default 'None' means that an exception
                        will be raised in such case.
        :return: element from YAML file (dict, array, string...)
        """
        tokens = path.split("|")
        current = self.config

        try:
            for token in tokens:
                current = current[token]
        except:
            if default is not None:
                return default
            else:
                raise KeyNotFoundException

        if isinstance(current, ScalarString):
            return str(current)
        else:
            return current

    def get_group_config(self, group) -> dict:
        pass

    def get_project_config(self, group_and_project) -> dict:
        pass

    def get_common_config(self) -> dict:
        """
        :return: literal common configuration or empty dict if not defined
        """
        common_config = self.get("projects_and_groups|*", {})
        if common_config:
            section_name = "*"
            self.validate_break_inheritance_flag(common_config, section_name)
        return common_config

    def is_project_skipped(self, project) -> bool:
        pass

    def is_group_skipped(self, group) -> bool:
        pass

    def is_skipped(self, an_array: list, item: str) -> bool:
        pass

    @staticmethod
    def validate_break_inheritance_flag(config, section_name, parent_key=""):
        for key, value in config.items():
            if "inherit" == key:
                parent_key_description = (
                    ' under key "' + parent_key + '"' if parent_key else ""
                )
                fatal(
                    f'The inheritance-break flag set in "{section_name}"{parent_key_description} is invalid\n'
                    f"because it has no higher level setting to inherit from.\n",
                    exit_code=EXIT_INVALID_INPUT,
                )
            elif type(value) is CommentedMap:
                ConfigurationCore.validate_break_inheritance_flag(
                    value, section_name, key
                )

    @staticmethod
    def merge_configs(more_general_config, more_specific_config) -> dict:
        """
        :return: merge more general config with more specific configs.
                 More specific config values take precedence over more general ones.
        """

        more_general_config = deepcopy(more_general_config)
        more_specific_config = deepcopy(more_specific_config)

        merged_dict = merge({}, more_general_config, more_specific_config)

        def break_inheritance(specific_config, parent_key=""):
            for key, value in specific_config.items():
                if "inherit" == key:
                    if not value:
                        replace_config_sections(
                            merged_dict, parent_key, specific_config
                        )
                        break
                    elif value:
                        fatal(
                            f"Cannot set the inheritance break flag with true\n",
                            exit_code=EXIT_INVALID_INPUT,
                        )
                elif type(value) is CommentedMap:
                    break_inheritance(value, key)

        def replace_config_sections(merged_config, specific_key, specific_config):
            for key, value in merged_config.items():
                if specific_key == key:
                    del specific_config["inherit"]
                    merged_config[key] = specific_config
                    break
                elif type(value) is CommentedMap:
                    replace_config_sections(value, specific_key, specific_config)

        break_inheritance(more_specific_config)

        return dict(merged_dict)


class ConfigFileNotFoundException(Exception):
    pass


class ConfigInvalidException(Exception):
    def __init__(self, underlying: Exception):
        self.underlying = underlying


class KeyNotFoundException(Exception):
    pass



import pytest

from gitlabform.gitlab.core import NotFoundException
from gitlabform.gitlab import AccessLevel
from tests.acceptance import (
    run_gitlabform,
    get_gitlab,
)


gl = get_gitlab()


class TestBranches:
    @pytest.mark.skipif(
        gl.has_no_license(), reason="this test requires a GitLab license (Paid/Trial)"
    )
    def test__code_owners_approval(self, gitlab, group_and_project, branch):

        try:
            branch_access_levels = gitlab.get_branch_access_levels(
                group_and_project, branch
            )
            assert branch_access_levels["code_owner_approval_required"] is False
        except NotFoundException:
            # this is fine, the branch may not be protected at all yet
            pass

        protect_branch_with_code_owner_approval_required = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              {branch}:
                protected: true
                push_access_level: maintainer
                merge_access_level: developer
                unprotect_access_level: maintainer
                code_owner_approval_required: true
        """

        run_gitlabform(
            protect_branch_with_code_owner_approval_required, group_and_project
        )

        branch_access_levels = gitlab.get_branch_access_levels(
            group_and_project, branch
        )
        assert branch_access_levels["code_owner_approval_required"] is True

    @pytest.mark.skipif(
        gl.has_no_license(), reason="this test requires a GitLab license (Paid/Trial)"
    )
    def test__allow_user_ids(
        self,
        gitlab,
        group_and_project,
        branch,
        make_user,
    ):

        user_allowed_to_push = make_user(AccessLevel.DEVELOPER)
        user_allowed_to_merge = make_user(AccessLevel.DEVELOPER)
        user_allowed_to_push_and_merge = make_user(AccessLevel.DEVELOPER)

        config_with_user_ids = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              {branch}:
                protected: true
                allowed_to_push:
                  - user_id: {user_allowed_to_push.id}
                  - access_level: {AccessLevel.NO_ACCESS.value}
                  - user: {user_allowed_to_push_and_merge.name}
                allowed_to_merge:
                  - access_level: {AccessLevel.DEVELOPER.value} 
                  - user_id: {user_allowed_to_merge.id}
                  - user: {user_allowed_to_push_and_merge.name}
        """

        run_gitlabform(config_with_user_ids, group_and_project)

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            _,
        ) = gitlab.get_only_branch_access_levels(group_and_project, branch)

        assert push_access_levels == [AccessLevel.NO_ACCESS.value]
        assert merge_access_levels == [AccessLevel.DEVELOPER.value]
        assert push_access_user_ids == sorted(
            [
                user_allowed_to_push.id,
                user_allowed_to_push_and_merge.id,
            ]
        )
        assert merge_access_user_ids == sorted(
            [
                user_allowed_to_merge.id,
                user_allowed_to_push_and_merge.id,
            ]
        )

    @pytest.mark.skipif(
        gl.has_no_license(), reason="this test requires a GitLab license (Paid/Trial)"
    )
    def test__allow_more_than_one_user_by_ids(
        self,
        gitlab,
        group_and_project,
        branch,
        make_user,
    ):
        first_user = make_user(AccessLevel.DEVELOPER)
        second_user = make_user(AccessLevel.DEVELOPER)
        third_user = make_user(AccessLevel.DEVELOPER)

        config_with_more_user_ids = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              {branch}:
                protected: true
                allowed_to_push:
                  - access_level: {AccessLevel.MAINTAINER.value} 
                  - user_id: {first_user.id}
                  - user_id: {second_user.id}
                  - user: {third_user.name}
                allowed_to_merge:
                  - access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(config_with_more_user_ids, group_and_project)

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            _,
        ) = gitlab.get_only_branch_access_levels(group_and_project, branch)

        assert push_access_levels == [AccessLevel.MAINTAINER.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]
        assert push_access_user_ids == sorted(
            [
                first_user.id,
                second_user.id,
                third_user.id,
            ]
        )
        assert merge_access_user_ids == []

    def test__protect_and_unprotect(self, gitlab, group_and_project, branch):

        config_protect_branch = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              {branch}:
                protected: true
                push_access_level: {AccessLevel.NO_ACCESS.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(config_protect_branch, group_and_project)

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, branch)
        assert push_access_levels == [AccessLevel.NO_ACCESS.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

        config_unprotect_branch = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              {branch}:
                protected: false
                push_access_level: {AccessLevel.NO_ACCESS.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(config_unprotect_branch, group_and_project)

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, branch)
        assert push_access_levels is None
        assert merge_access_levels is None
        assert push_access_user_ids is None
        assert merge_access_user_ids is None
        assert unprotect_access_level is None

    def test__config_with_access_level_names(self, gitlab, group_and_project, branch):

        config_with_access_levels_names = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              {branch}:
                protected: true
                push_access_level: no_access        # note "_" or " " and the various
                merge_access_level: Developer       # case in each line. it should not
                unprotect_access_level: MAINTAINER  # matter as we allow any case.
        """

        run_gitlabform(config_with_access_levels_names, group_and_project)

        (
            push_access_level,
            merge_access_level,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, branch)
        assert push_access_level == [AccessLevel.NO_ACCESS.value]
        assert merge_access_level == [AccessLevel.DEVELOPER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        assert unprotect_access_level is AccessLevel.MAINTAINER.value



from abc import ABC, abstractmethod
from types import SimpleNamespace

from cli_ui import fatal
from yamlpath import Processor
from yamlpath.exceptions import YAMLPathException
from yamlpath.wrappers import ConsolePrinter

from gitlabform import EXIT_INVALID_INPUT
from gitlabform.configuration import Configuration
from gitlabform.gitlab import AccessLevel


class ConfigurationTransformer(ABC):
    @classmethod
    @abstractmethod
    def transform(cls, configuration: Configuration):
        pass


class AccessLevelsTransformer(ConfigurationTransformer):
    """
    Internally the app supports only numeric access levels, but for user-friendliness
    this class allows providing them as strings and transforms these strings into
    the appropriate numbers.
    """

    @classmethod
    def transform(cls, configuration: Configuration):
        logging_args = SimpleNamespace(quiet=False, verbose=False, debug=False)
        log = ConsolePrinter(logging_args)

        processor = Processor(log, configuration.config)

        # [.!<100] effectively means that the value is non-numerical
        paths_to_hashes = [
            # # branches, old syntax
            "**.push_access_level[.!<100]",
            "**.merge_access_level[.!<100]",
            "**.unprotect_access_level[.!<100]",
            # members & group members
            "**.access_level[.!<100]",
            "**.group_access[.!<100]",
            # old syntax
            "**.group_access_level[.!<100]",
            # tags
            "**.create_access_level[.!<100]",
        ]

        for path in paths_to_hashes:
            try:
                for node_coordinate in processor.get_nodes(path):
                    try:
                        access_level_string = str(node_coordinate.node)
                        node_coordinate.parent[
                            node_coordinate.parentref
                        ] = AccessLevel.get_value(access_level_string)
                    except KeyError:
                        fatal(
                            f"Configuration string '{access_level_string}' is not one of the valid access levels:"
                            f" {', '.join(AccessLevel.get_canonical_names())}",
                            exit_code=EXIT_INVALID_INPUT,
                        )
            except YAMLPathException:
                pass

        # there are different than the above, as they are elements of arrays
        # so we need different search query and an extra condition for
        # transformation
        paths_to_arrays = [
            # # branches, new GitLab Premium syntax
            "**.allowed_to_push.*.[access_level!<100]",
            "**.allowed_to_merge.*.[access_level!<100]",
            "**.allowed_to_unprotect.*.[access_level!<100]",
        ]

        for path in paths_to_arrays:
            try:
                for node_coordinate in processor.get_nodes(path):
                    if node_coordinate.parentref == "access_level":
                        try:
                            access_level_string = str(node_coordinate.node)
                            node_coordinate.parent[
                                node_coordinate.parentref
                            ] = AccessLevel.get_value(access_level_string)
                        except KeyError:
                            fatal(
                                f"Configuration string '{access_level_string}' is not one of the valid access levels:"
                                f" {', '.join(AccessLevel.get_canonical_names())}",
                                exit_code=EXIT_INVALID_INPUT,
                            )
            except YAMLPathException:
                pass



from cli_ui import fatal

from gitlabform import EXIT_INVALID_INPUT
from gitlabform.configuration.core import KeyNotFoundException
from gitlabform.configuration.projects_and_groups import ConfigurationProjectsAndGroups


class ConfigurationCaseInsensitiveProjectsAndGroups(ConfigurationProjectsAndGroups):
    def __init__(self, config_path=None, config_string=None):
        super().__init__(config_path, config_string)
        self.find_almost_duplicates()

    def get_group_config(self, group) -> dict:
        """
        :param group: group/subgroup
        :return: configuration for this group/subgroup or empty dict if not defined,
                 ignoring the case
        """
        try:
            return self.get_case_insensitively(
                self.get("projects_and_groups"), f"{group}/*"
            )
        except KeyNotFoundException:
            return {}

    def get_project_config(self, group_and_project) -> dict:
        """
        :param group_and_project: 'group/project'
        :return: configuration for this project or empty dict if not defined,
                 ignoring the case
        """
        try:
            return self.get_case_insensitively(
                self.get("projects_and_groups"), group_and_project
            )
        except KeyNotFoundException:
            return {}

    def is_project_skipped(self, project) -> bool:
        """
        :return: if project is defined in the key with projects to skip,
                 ignoring the case
        """
        return self.is_skipped_case_insensitively(
            self.get("skip_projects", []), project
        )

    def is_group_skipped(self, group):
        """
        :return: if group is defined in the key with groups to skip,
                 ignoring the case
        """
        return self.is_skipped_case_insensitively(self.get("skip_groups", []), group)

    @staticmethod
    def get_case_insensitively(a_dict: dict, a_key: str):
        for dict_key in a_dict.keys():
            if dict_key.lower() == a_key.lower():
                return a_dict[dict_key]
        raise KeyNotFoundException()

    @staticmethod
    def is_skipped_case_insensitively(an_array: list, item: str) -> bool:
        """
        :return: if item is defined in the list to be skipped
        """
        item = item.lower()

        for list_element in an_array:
            list_element = list_element.lower()

            if list_element == item:
                return True

            if (
                list_element.endswith("/*")
                and item.startswith(list_element[:-2])
                and len(item) >= len(list_element[:-2])
            ):
                return True

        return False

    def find_almost_duplicates(self):

        # in GitLab groups and projects names are de facto case insensitive:
        # you can change the case of both name and path BUT you cannot create
        # 2 groups which names differ only with case and the same thing for
        # projects. therefore we cannot allow such entries in the config,
        # as they would be ambiguous.

        for path in [
            "projects_and_groups",
            "skip_groups",
            "skip_projects",
        ]:
            if self.get(path, 0):
                almost_duplicates = self._find_almost_duplicates(path)
                if almost_duplicates:
                    fatal(
                        f"There are almost duplicates in the keys of {path} - they differ only in case.\n"
                        f"They are: {', '.join(almost_duplicates)}\n"
                        f"This is not allowed as we ignore the case for group and project names.",
                        exit_code=EXIT_INVALID_INPUT,
                    )

    def _find_almost_duplicates(self, configuration_path):
        """
        Checks given configuration key and reads its keys - if it is a dict - or elements - if it is a list.
        Looks for items that are almost the same - they differ only in the case.
        :param configuration_path: configuration path, f.e. "group_settings"
        :return: list of items that have almost duplicates,
                 or an empty list if none are found
        """

        dict_or_list = self.get(configuration_path)
        if isinstance(dict_or_list, dict):
            items = dict_or_list.keys()
        else:
            items = dict_or_list

        items_with_lowercase_names = [x.lower() for x in items]

        # casting these to sets will deduplicate the one with lowercase names
        # lowering its cardinality if there were elements in it
        # that before lowering differed only in case
        if len(set(items)) != len(set(items_with_lowercase_names)):

            # we have some almost duplicates, let's find them
            almost_duplicates = []
            for first_item in items:
                occurrences = 0
                for second_item in items_with_lowercase_names:
                    if first_item.lower() == second_item.lower():
                        occurrences += 1
                        if occurrences == 2:
                            almost_duplicates.append(first_item)
                            break
            return almost_duplicates

        else:
            return []



from tests.acceptance import run_gitlabform, DEFAULT_README


class TestFilesTemplates:
    def test__default_variables(self, gitlab, group_and_project):

        config = (
            """
        projects_and_groups:
          """
            + group_and_project
            + """:
            files:
              "README.md":
                overwrite: true
                branches: all
                content: |
                  This is a text that contains the default variables: {{ group }}/{{ project }}
        """
        )

        run_gitlabform(config, group_and_project)

        file_content = gitlab.get_file(group_and_project, "main", "README.md")
        # TODO: the file_content should actually end with a new line, as this is how the original content above ends
        assert (
            file_content
            == f"This is a text that contains the default variables: "
            + group_and_project
        )

    def test__custom_variables(self, gitlab, group_and_project):

        config = (
            """
        projects_and_groups:
          """
            + group_and_project
            + """:
            files:
              "README.md":
                overwrite: true
                branches: all
                jinja_env:
                  foo: "fooz"
                  bar: "barz"
                content: |
                  This is a text that contains custom variables: {{ foo }}, {{ bar }}
        """
        )

        run_gitlabform(config, group_and_project)

        file_content = gitlab.get_file(group_and_project, "main", "README.md")
        # TODO: the file_content should actually end with a new line, as this is how the original content above ends
        assert (
            file_content == "This is a text that contains custom variables: fooz, barz"
        )

    def test__disabled_templating(self, gitlab, group_and_project):

        config = (
            """
        projects_and_groups:
          """
            + group_and_project
            + """:
            files:
              "README.md":
                overwrite: true
                branches: all
                template: no
                content: |
                  This is a text containing literals: {{ group }}/{{ project }}
        """
        )

        run_gitlabform(config, group_and_project)

        file_content = gitlab.get_file(group_and_project, "main", "README.md")
        assert (
            file_content
            == "This is a text containing literals: {{ group }}/{{ project }}\n"
        )



from tests.acceptance import (
    run_gitlabform,
)


def single_true(iterable):
    i = iter(iterable)
    # use the fact that thanks to iter() we will "consume"
    # all the elements of iterable up to the one that is True
    # so that there should be no other True in the rest
    return any(i) and not any(i)


class TestDeployKeys:
    def test__deploy_key_add(self, gitlab, group_and_project_for_function):

        # noinspection PyPep8
        key = "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABgQDydRMMBVOXfD76kjfI9wjxcSkJqdqr9222naDd6EGLuoDPjwRdowGJXTPLO88iFux2hWpROfT3PaP8s/yi0LkKSi2Fwn4hEc9u8pVoXQVhEwu367cLmy0XCo5lOmkfXaBSvLqb+Z6v9CozdJzmsllcTCK1DoYcGD8NPnQMMEETqbHzropjUjLA/to/zfI/mYVP86X15w+pw0DsUtspj3MmQBxPks4v2EAF7tUGLgqMHMr/z5bsWkm6yR4fv7rfjLoh10tXUY8WrVvzAZzWCs7fnP1qf5CCU7MlzggSIhzbwLn2DcYMBnFKGgx3H/VwJIvtmoIq8duedlUQ8zxKaSK8ziF/WQ8EtMW19qCrI8W+6vOgJpooDpBkPnSE+gsS+ANyWoXOJhgGukjPtphYqGTvDQAAbAMIeXB7QqDwq62UkgRSr5TC4pTVQrzlRTLxrnWMpYhpYy/3fCvYWDvuRFV8+IH6mlXoCrcMfh78oShmwkv8+A9/j9pBBBiFBZ2x6sM="

        config_add = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                key: {key}
                title: some_key
        """
        run_gitlabform(config_add, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "some_key" for key in deploy_keys])

    def test__deploy_key_add_delete(self, gitlab, group_and_project_for_function):

        # noinspection PyPep8
        key = "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABgQDydRMMBVOXfD76kjfI9wjxcSkJqdqr9222naDd6EGLuoDPjwRdowGJXTPLO88iFux2hWpROfT3PaP8s/yi0LkKSi2Fwn4hEc9u8pVoXQVhEwu367cLmy0XCo5lOmkfXaBSvLqb+Z6v9CozdJzmsllcTCK1DoYcGD8NPnQMMEETqbHzropjUjLA/to/zfI/mYVP86X15w+pw0DsUtspj3MmQBxPks4v2EAF7tUGLgqMHMr/z5bsWkm6yR4fv7rfjLoh10tXUY8WrVvzAZzWCs7fnP1qf5CCU7MlzggSIhzbwLn2DcYMBnFKGgx3H/VwJIvtmoIq8duedlUQ8zxKaSK8ziF/WQ8EtMW19qCrI8W+6vOgJpooDpBkPnSE+gsS+ANyWoXOJhgGukjPtphYqGTvDQAAbAMIeXB7QqDwq62UkgRSr5TC4pTVQrzlRTLxrnWMpYhpYy/3fCvYWDvuRFV8+IH6mlXoCrcMfh78oShmwkv8+A9/j9pBBBiFBZ2x6sM="

        config_add = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                key: {key}
                title: some_key
        """
        run_gitlabform(config_add, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "some_key" for key in deploy_keys])

        # noinspection PyPep8
        config_delete = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                title: some_key
                delete: true
        """
        run_gitlabform(config_delete, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert not any([key["title"] == "some_key" for key in deploy_keys])

    def test__deploy_key_add_delete_readd(self, gitlab, group_and_project_for_function):

        # noinspection PyPep8
        key = "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABgQDydRMMBVOXfD76kjfI9wjxcSkJqdqr9222naDd6EGLuoDPjwRdowGJXTPLO88iFux2hWpROfT3PaP8s/yi0LkKSi2Fwn4hEc9u8pVoXQVhEwu367cLmy0XCo5lOmkfXaBSvLqb+Z6v9CozdJzmsllcTCK1DoYcGD8NPnQMMEETqbHzropjUjLA/to/zfI/mYVP86X15w+pw0DsUtspj3MmQBxPks4v2EAF7tUGLgqMHMr/z5bsWkm6yR4fv7rfjLoh10tXUY8WrVvzAZzWCs7fnP1qf5CCU7MlzggSIhzbwLn2DcYMBnFKGgx3H/VwJIvtmoIq8duedlUQ8zxKaSK8ziF/WQ8EtMW19qCrI8W+6vOgJpooDpBkPnSE+gsS+ANyWoXOJhgGukjPtphYqGTvDQAAbAMIeXB7QqDwq62UkgRSr5TC4pTVQrzlRTLxrnWMpYhpYy/3fCvYWDvuRFV8+IH6mlXoCrcMfh78oShmwkv8+A9/j9pBBBiFBZ2x6sM="

        config_add = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                key: {key}
                title: some_key
        """
        run_gitlabform(config_add, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "some_key" for key in deploy_keys])

        # noinspection PyPep8
        config_delete = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                title: some_key
                delete: true
        """
        run_gitlabform(config_delete, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert not any([key["title"] == "some_key" for key in deploy_keys])

        run_gitlabform(config_add, group_and_project_for_function)
        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "some_key" for key in deploy_keys])

    def test__deploy_key_add_delete_readd_under_different_name(
        self, gitlab, group_and_project_for_function
    ):

        # noinspection PyPep8
        key = "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABgQDydRMMBVOXfD76kjfI9wjxcSkJqdqr9222naDd6EGLuoDPjwRdowGJXTPLO88iFux2hWpROfT3PaP8s/yi0LkKSi2Fwn4hEc9u8pVoXQVhEwu367cLmy0XCo5lOmkfXaBSvLqb+Z6v9CozdJzmsllcTCK1DoYcGD8NPnQMMEETqbHzropjUjLA/to/zfI/mYVP86X15w+pw0DsUtspj3MmQBxPks4v2EAF7tUGLgqMHMr/z5bsWkm6yR4fv7rfjLoh10tXUY8WrVvzAZzWCs7fnP1qf5CCU7MlzggSIhzbwLn2DcYMBnFKGgx3H/VwJIvtmoIq8duedlUQ8zxKaSK8ziF/WQ8EtMW19qCrI8W+6vOgJpooDpBkPnSE+gsS+ANyWoXOJhgGukjPtphYqGTvDQAAbAMIeXB7QqDwq62UkgRSr5TC4pTVQrzlRTLxrnWMpYhpYy/3fCvYWDvuRFV8+IH6mlXoCrcMfh78oShmwkv8+A9/j9pBBBiFBZ2x6sM="

        config_add = f"""
            projects_and_groups:
              {group_and_project_for_function}:
                deploy_keys:
                  foobar:
                    key: {key}
                    title: a_title
            """
        run_gitlabform(config_add, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "a_title" for key in deploy_keys])

        # noinspection PyPep8
        config_delete = f"""
            projects_and_groups:
              {group_and_project_for_function}:
                deploy_keys:
                  foobar:
                    title: a_title
                    delete: true
            """
        run_gitlabform(config_delete, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert not any([key["title"] == "a_title" for key in deploy_keys])

        # noinspection PyPep8
        config_readd = f"""
            projects_and_groups:
              {group_and_project_for_function}:
                deploy_keys:
                  foobar:
                    key: {key}
                    title: another_title
            """
        run_gitlabform(config_readd, group_and_project_for_function)
        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "another_title" for key in deploy_keys])

    def test__deploy_key_update_title(self, gitlab, group_and_project_for_function):

        key = "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABgQCz7QLHyLval1TS+MvLN9QwLi5It7f74BQWi7d5gxBKA7BaGxebOc3AmLF9B/XkBfSNLOEhKO6NZrq07U4C2c3j29IvHy1oAk/cg6YmDL6U7d9lxBxYHC/SsnN70LekGbjHhaMofNvfnDOpS7n5MvJjY42ovxag+SRAOhtp541no6R/Oj9OoW6Y3AtX59HIcP1JvGTx/Ohb8OUwOXbfIDxrqR4qt1kiSPLAAC8wLNNDTUYs89TVAqyvzuXXindmcXosVcEQ6EYzHPin66ge1rAYJfhBqei1tmS0OUrW4awzQxddBqrwVBq0occUrbkjJNRfEjHjYR2GDutk8bP/kZ7cJ3RMU7bCh6CNGvApN7BysSfFcS19/18BCVjWFWbZSHGoaB0DDjl9R5s7RzMuBvNULt4yTEQUOQKdmteJY6RxKApxiCglu8I+8fIzL75iDNejZ+UlXj1SnIfe2BrzR/EN2FAGubq2SLmKLGLSGk3lkwDBdPfNMjYJG9bdaeHt2mc="

        config = f"""
            projects_and_groups:
              {group_and_project_for_function}:
                deploy_keys:
                  foobar:
                    key: {key}
                    title: title_before
            """
        run_gitlabform(config, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "title_before" for key in deploy_keys])

        config = f"""
            projects_and_groups:
              {group_and_project_for_function}:
                deploy_keys:
                  # because it's the title that defines the key, we cannot immediately change key's title in the config.
                  # we need two element config - 1. delete it under old title, 2. add it under new title
                  foo:
                    title: title_before
                    delete: true
                  bar:
                    key: {key}
                    title: title_after
            """
        run_gitlabform(config, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "title_after" for key in deploy_keys])
        assert not any([key["title"] == "title_before" for key in deploy_keys])

    def test__deploy_key_update_value(self, gitlab, group_and_project_for_function):

        # noinspection PyPep8
        key_before = "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABgQCz7QLHyLval1TS+MvLN9QwLi5It7f74BQWi7d5gxBKA7BaGxebOc3AmLF9B/XkBfSNLOEhKO6NZrq07U4C2c3j29IvHy1oAk/cg6YmDL6U7d9lxBxYHC/SsnN70LekGbjHhaMofNvfnDOpS7n5MvJjY42ovxag+SRAOhtp541no6R/Oj9OoW6Y3AtX59HIcP1JvGTx/Ohb8OUwOXbfIDxrqR4qt1kiSPLAAC8wLNNDTUYs89TVAqyvzuXXindmcXosVcEQ6EYzHPin66ge1rAYJfhBqei1tmS0OUrW4awzQxddBqrwVBq0occUrbkjJNRfEjHjYR2GDutk8bP/kZ7cJ3RMU7bCh6CNGvApN7BysSfFcS19/18BCVjWFWbZSHGoaB0DDjl9R5s7RzMuBvNULt4yTEQUOQKdmteJY6RxKApxiCglu8I+8fIzL75iDNejZ+UlXj1SnIfe2BrzR/EN2FAGubq2SLmKLGLSGk3lkwDBdPfNMjYJG9bdaeHt2mc="
        key_after = "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAICRgWbMGu+88gPVPEIAHQdG958oTzSFmuBnwmC0hunyp"

        config = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                key: {key_before}
                title: a_key
        """
        run_gitlabform(config, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert any([key["key"] == key_before for key in deploy_keys])

        config = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                key: {key_after}
                title: a_key
        """
        run_gitlabform(config, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert not any([key["key"] == key_before for key in deploy_keys])
        assert any([key["key"] == key_after for key in deploy_keys])
